# to run the experiments

# make sure the setwd() is where main_simulation.R lies.

source("../helpers.R")
library("parallel")
library("ranger")
library(mvtnorm)
library(naniar)
library(MASS)
library(stats)
library(fields)



n.cores <- 1

missing.mechs <- c("MAR","MCAR")
cases <- 1:8
nsim <- 300
n <-500
p <-10
r <-0.35

## PKLM choices
num.proj <- 100
nrep <- 30
refit<-T
num.trees.per.proj <- 200
min.node.size <- 10
size.resp.set <- 2 #"random"

runEnergyTest <-FALSE # This is always set to False, since the Q test is not included here.
runlittleTest <- TRUE
runJJTest <- TRUE
level <- 0.05
kind <- "mean"

marmech <- "basic" #"random"



seed <- sample(c(0:2000),1)
set.seed(seed)



for (mech in missing.mechs){
  for (case in cases){
    
    # For testing
    # mech <-type<-"MCAR"
    # case=5
    # num.trees.per.proj = 10
    
    
    pvals <- runSims(nsim = nsim, type=mech,runEnergyTest=runEnergyTest, runJJTest=runJJTest, runlittleTest=runlittleTest, case = case, n = n, p = p, r = r,
                     num.proj = num.proj, num.trees.per.proj = num.trees.per.proj,
                     min.node.size = min.node.size, marmech = marmech,
                     level=level,nrep = nrep, n.cores=n.cores, kind=kind, size.resp.set=size.resp.set, refit=refit)
    
    print(paste("Case", case, "done"))
    
    filename = paste0("case_", case,
                      "_mech_", mech,
                      "_energy_",runEnergyTest,
                      "_nrep_",nrep,
                      "_numproj_", num.proj,
                      "_n_",n,
                      "_r_",r,
                      "_p_",p,
                      "_nsim_",nsim,
                      "_level_",level,
                      "_minnodesize_", min.node.size,
                      "_numtreesperproj_", num.trees.per.proj,
                      "_size.resp.set_",size.resp.set,
                      "_kind_", kind
    )
    
    assign(filename, pvals)
    save(pvals, file=paste(filename, ".Rda",sep=""))
    
    
  }
  
}

