


gendist<-function(n,p,dist, ...){

  if (dist=="Gaussian"){
    Z <- MASS::mvrnorm(n = n, ...)

  }

  if (dist=="t"){


    arguments <- list(...)
    Sigma<-unlist(arguments$Sigma)
    df<-unlist(arguments$df)
    mu<-unlist(arguments$mu)

    Z <-  mu + rmvt(n, sigma=Sigma, df=df)

  }

  if (dist=="uniform"){

    Z0<-matrix(runif(n*p), nrow = n, ncol=p)

    arguments <- list(...)
    Sigma<-unlist(arguments$Sigma)
    Sigmasvd<-svd(Sigma)
    ## Check!
    Sigmasqrt<-Sigmasvd$u%*%diag(sqrt(Sigmasvd$d))%*%t(Sigmasvd$v)

    Z<-Z0%*%Sigmasqrt

  }

  if (dist=="W"){

    Z0 <- MASS::mvrnorm(n = n, ...)
    Z <- Z0 + 0.1*Z0^3

  }

  if (dist=="Weibull"){

    Z<-matrix(rweibull(n*p, ...), nrow = n, ncol=p)


  }

  return(Z)

}


genData_real <- function(dataset= "dengue", datapath=NULL){
  
  
  path0<-getwd()
  
  if (!is.null(datapath)){
    
    setwd(datapath)
  }
  
  if (dataset == "dengue"){
    # description: dengue fever in africa.
    # https://www.picostat.com/dataset/r-dataset-package-daag-dengue
    dat <- read.csv(file="dataset_dengue.csv")
    X <- as.matrix(data.frame(na.omit(dat)))
    colnames(X) <-NULL
    
  }else if (dataset== "Boston"){
    X <- as.matrix(Boston) ## For the moment: take out categorical variables
    colnames(X) <- NULL
    
  }else if (dataset == "network"){
    nv <- vcount(fblog)
    X <- as_adjacency_matrix(fblog)
    colnames(X) <- NULL
    rownames(X) <- NULL
    
  }else if (dataset == "iris"){
    X <- iris
    X <- X[,-ncol(X)]
    X <- as.matrix(X)
    colnames(X) <- NULL
    
    
  }else if (dataset == "wine"){
    data("wine")
    X <- wine
    X$Class <- NULL
    X <- as.matrix(X)
    colnames(X) <- NULL
    
    
  }else if(dataset == "wisconsin"){
    
    dat <- read.csv(file="wisconsin.csv")
    dat<-dat[,-c(1,2,33)]
    X <- as.matrix(dat)
    colnames(X) <-NULL
    
  }else if(dataset == "wine.quality.red"){ #https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/
    dat <- read.csv(file="winequality-red.csv", sep=";")
    dat$quality<-NULL
    X <- as.matrix(dat)
    colnames(X) <-NULL
    
    
  }else if(dataset == "wine.quality.white"){ #https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/
    dat <- read.csv(file="winequality-white.csv", sep=";")
    dat$quality<-NULL
    X <- as.matrix(dat)
    colnames(X) <- NULL
    
    
  }else if(dataset == "yacht"){ #https://archive.ics.uci.edu/ml/machine-learning-databases/00243/
    X <- read.csv(file="yacht_hydrodynamics.data", sep="", header=F)
    X <- as.matrix(X)
    colnames(X) <-NULL
    
  }else if(dataset == "yeast"){ #https://archive.ics.uci.edu/ml/machine-learning-databases/yeast/
    X <- read.csv(file="yeast.data", sep="", header=F)
    X <- X[,-c(1, ncol(X))]
    X <- as.matrix(X)
    colnames(X) <-NULL
    
    
  }else if(dataset == "DNA"){
    data("DNA")
    dat <- DNA
    X <- as.matrix(model.matrix(~., dat))
    X<-X[,-1]
    colnames(X) <-NULL
    
  }else if(dataset == "airfoil"){
    X <- read.table(file="airfoil_self_noise.dat")
    colnames(X) <-NULL
    X <-as.matrix(X)
    
  }else if (dataset =="blood.transfusion"){
    X <- read.csv(file="transfusion.data")
    X<- X[,-ncol(X)]
    X <- as.matrix(X)
    colnames(X)<- NULL
    
    
  }else if (dataset == "concrete.slump"){
    X <- read.csv(file="slump_test.data")
    X <- as.matrix(X)
    X <- X[,-1]
    colnames(X)<- NULL
    
  }else if (dataset == "connectionist.bench.sonar"){
    
    X <- read.csv(file="sonar.all-data", header=F)
    X <- X[,-ncol(X)]
    X <- as.matrix(X)
    colnames(X)<- NULL
    
    #
  }else if (dataset == "connectionist.bench.vowel"){
    X <- read.csv(file="vowel-context.data", sep="", header=F)
    X <- X[,-c(1,2,3, ncol(X))]
    X <- as.matrix(X)
    colnames(X)<- NULL
    
    
  }else if (dataset == "ecoli"){
    X <- read.csv(file="ecoli.data", sep="", header=F)
    X <- X[,-c(1, ncol(X))]
    X <- as.matrix(X)
    X <- X[,-7]# hihgly cor with var8
    X <- X[,-4]
    colnames(X)<- NULL
    
    
  }else if (dataset=="glass"){
    X <- read.csv(file="glass.data",header=F)
    X <- X[,-c(1, ncol(X))]
    colnames(X)<- NULL
    
    
    
  }else if (dataset=="ionosphere"){ #https://archive.ics.uci.edu/ml/machine-learning-databases/ionosphere/
    X <- read.csv(file="ionosphere.data",header=F)
    X <- X[,-c(1,2, ncol(X))]
    X <- as.matrix(X)
    colnames(X)<- NULL
    
    
  }else if (dataset=="libras"){ #https://archive.ics.uci.edu/ml/machine-learning-databases/libras/
    X <- read.csv(file="movement_libras.data", header=F)
    X <- X[,-ncol(X)]
    X <- as.matrix(X)
    colnames(X)<- NULL
    
    
  }else if (dataset=="parkinsons"){ #https://archive.ics.uci.edu/ml/machine-learning-databases/parkinsons/
    X <- read.csv(file="parkinsons.data", header=T)
    X$status <- NULL
    X$name <- NULL
    X <- as.matrix(X)
    colnames(X)<- NULL
    
  }else if (dataset=="planning.relax"){ #https://archive.ics.uci.edu/ml/machine-learning-databases/00230/
    X <- read.csv("plrx.txt", sep="", header=F)
    X <- X[,-ncol(X)]
    X <- as.matrix(X)
    colnames(X)<- NULL
    
    
    #
  }else if (dataset=="seeds"){ #https://archive.ics.uci.edu/ml/machine-learning-databases/00236/
    X <- read.csv("seeds_dataset.txt", sep="", header=F)
    X <- X[,-ncol(X)]
    X <- as.matrix(X)
    colnames(X)<- NULL
    
    
    
  }else if (dataset=="climate.model.crashes"){
    X <- read.table(file="pop_failures.dat", header = TRUE)
    X <- X[,-c(1,ncol(X))]
    X<-as.matrix(X)
    colnames(X)<- NULL
    
    
    #
  }else if (dataset == "concrete.compression"){
    X <- read_excel("Concrete_Data.xls")
    X <- as.matrix(X)
    colnames(X)<- NULL
    
  }else if (dataset == "CASchools"){
    data("CASchools")
    X <- CASchools
    X <- X[,-c(1,2,3,4)]
    X <- as.matrix(X)
    colnames(X) <- NULL
    
    
  }
  setwd(path0)
  
  return(X)
  
}




genMask <- function(X, n, p, r, type = "MCAR") {
  if (type == "MCAR") {
    M <- matrix(0, ncol = p, nrow = n)
    M <- apply(M, 2, function(x) ifelse(runif(n)<= 1-r^{1/p}, 1, x))
    return(M)
  } else if (type == "MAR") {
    M <- matrix(0, ncol = p, nrow = n)
    # Randomly scatter missingness such that we have r complete cases
    M[,-1] <- apply(M[,-1, drop=F], 2, function(x) ifelse(runif(n)<= 1-r^{1/(p-1)}, 1, x))
    # Indices of incomplete rows (missing group)
    ids.inc <- which(sapply(1:nrow(M), function(i) any((M[i,]==1))))
    ind <- rep(NA, nrow(X))
    # Indices of complete rows (complete group)
    full.indices <- setdiff(1:nrow(X), ids.inc)
    for (i in 1:nrow(X)){
      if (X[i,1]<=mean(X[,1],na.rm=TRUE)) {
        group <- sample(1:2, 1,prob=c(1/6,5/6)) #2 means sample from the missing group
        if(group==2){
          if(length(ids.inc)>=1){
            if(length(ids.inc)==1){
              ind[i] <-  ids.inc
              ids.inc <- numeric()
            }else{
              ind[i] <-  sample(ids.inc, size = 1)
              ids.inc <- setdiff(ids.inc, ind[i])
            }

          }else{
            if(length(full.indices)==1){
              ind[i] <-  full.indices
              full.indices<- numeric()
            }else{
              ind[i] <-  sample(full.indices, size = 1)
              full.indices <- setdiff(full.indices, ind[i])
            }
          }
        }else{
          if(length(full.indices)>=1){
            if(length(full.indices)==1){
              ind[i] <-  full.indices
              full.indices <- numeric()
            }else{
              ind[i] <-  sample(full.indices, size = 1)
              full.indices <- setdiff(full.indices, ind[i])
            }

          }else{
            if(length(ids.inc)==1){
              ind[i] <-  ids.inc
              ids.inc <- numeric()
            }else{
              ind[i] <-  sample(ids.inc, size = 1)
              ids.inc <- setdiff(ids.inc, ind[i])
            }

          }
        }
      }else {
        group <- sample(1:2, 1,prob=c(5/6,1/6)) #2 means sample from the missing group
        if(group==2){
          if(length(ids.inc)>=1){
            if(length(ids.inc)==1){
              ind[i] <-  ids.inc
              ids.inc <- numeric()
            }else{
              ind[i] <-  sample(ids.inc, size = 1)
              ids.inc <- setdiff(ids.inc, ind[i])
            }

          }else{
            if(length(full.indices)==1){
              ind[i] <-  full.indices
              full.indices<- numeric()
            }else{
              ind[i] <-  sample(full.indices, size = 1)
              full.indices <- setdiff(full.indices, ind[i])
            }
          }
        }else{
          if(length(full.indices)>=1){
            if(length(full.indices)==1){
              ind[i] <-  full.indices
              full.indices <- numeric()
            }else{
              ind[i] <-  sample(full.indices, size = 1)
              full.indices <- setdiff(full.indices, ind[i])
            }

          }else{
            if(length(ids.inc)==1){
              ind[i] <-  ids.inc
              ids.inc <- numeric()
            }else{
              ind[i] <-  sample(ids.inc, size = 1)
              ids.inc <- setdiff(ids.inc, ind[i])
            }

          }
        }

      }
    }


    M <- M[ind,]
    return(M)
  }
}

genMAR_original <- function(X, miss.p=0.1) {
  p <- ncol(X)
  n <- nrow(X)

  bound.X=quantile(X[,1],(1-miss.p)^(p/(p-1)))
  miss.sample=cbind(0,X[,-p])
  miss.ind=matrix(as.numeric(miss.sample<bound.X),nr=n)
  X.NA <- X
  X.NA[miss.ind==0] <- NA
  return(X.NA)
}

genMCAR_original <- function(X, miss.p=0.1) {
  p <- ncol(X)
  n <- nrow(X)

  miss.ind=matrix(0,nr=n,nc=p)
  flag=1:n
  while(length(flag)>0){
    miss.sample=runif(length(flag)*p)
    miss.ind[flag,]=matrix(as.numeric(miss.sample<(1-miss.p)),nc=p,byrow=TRUE)   #determine the missing mechanism(observed=1,missing=0)
    flag=which(apply(miss.ind,1,max)==0)
  }
  X.NA <- X
  X.NA[miss.ind==0] <- NA
  return(X.NA)
}


# prob truncation
truncProb <- function(p) {
  pmax(pmin(p, 1-10^{-9}),10^{-9})
}

# generate the statistics U
genU <- function(st, lab) {

  # OOB preds
  preds <- st$predictions

  # number of actual classes
  ui <- rep(NA, ncol(preds))

  for (i in 1:length(ui)) {

    ind <- colnames(preds)[i]
    p <- preds[lab==ind,i]
    p <- truncProb(p)
    lab2 <- lab[is.finite(p)]
    p <- p[is.finite(p)]

    p.a <- preds[lab!=ind,i]
    p.a <- truncProb(p.a)
    lab2.a <- lab[is.finite(p.a)]
    p.a <- p.a[is.finite(p.a)]


    ui[i] <- mean(log(p/(1-p))) - mean(log(p.a/(1-p.a)))
  }

  return(mean(ui, na.rm=TRUE))
}






pvalsDecision <- function(pvals, gamma = 0.1, level = 0.05) {
  return(as.numeric((1/gamma)*quantile(pvals, gamma, na.rm=TRUE) <= level))
}



pvalsDecision_adapt <- function(pvals, level = 0.05) {

  objective <- function(w) {min(1, (1/w)*quantile(pvals, w))}
  min.optim <- optim(0, fn=objective, lower=level, upper=1, method="L-BFGS-B")$value

  grid.gamma = seq(level,0.1,0.00005)
  q.gamma <- sapply(grid.gamma, function(g) min(1, (1/g)*quantile(pvals, g)))

  min.q.gamma <- min(min(q.gamma), min.optim)

  p.adj <- min(1, (1-log(level))*min.q.gamma)
  return(p.adj)
}






edist <- function(X, Y) {
  A <- mean(apply(X, 1, function(x) apply(Y, 1, function(y) sqrt(sum((x-y)^2)))))
  B <- mean(apply(X, 1, function(x) apply(X, 1, function(y) sqrt(sum((x-y)^2)))))
  C <- mean(apply(Y, 1, function(x) apply(Y, 1, function(y) sqrt(sum((x-y)^2)))))
  return(2*A-B-C)
}

gfunc <- function(X, Y) {
  A <- mean(apply(X, 1, function(x) apply(Y, 1, function(y) sqrt(sum((x-y)^2)))))
  return(A)
}

runSims <- function(nsim, type, runPKLMtest=TRUE, runEnergyTest=FALSE, runJJTest=FALSE, runlittleTest=FALSE, case, n = 200,
                    p = 10, r = 0.65,n.cores,size.resp.set,pmiss, marmech ,...) {
  mu <-rep(0,p)
  Sigma <- 0.7* matrix(1, nrow=p, ncol=p) + 0.3*diag(p)

  pvals.PKLMtest <- rep(NA, nsim)
  pvals.energyTest <- rep(NA, nsim)
  
  args<-list(...)
  
  if ("refit" %in% names(args)){
    refit<-args$refit
  }else{
      refit<-F
  }
  
  if ("datapath" %in% names(args)){
    datapath<-args$datapath
  }else{
    datapath<-NULL
  }
  
  
  pvals <- mclapply(1:nsim, FUN = function(i){

    if(case ==1){
      #(i) Standard Normal (N)
      Z<-gendist(n=n,p=p, dist="Gaussian", Sigma=diag(p), mu=rep(0,p))
      if(marmech =="basic"){
        M <- genMask(X = Z, n = n, p = p, r = r, type=  type)
        Z.NA <- Z
        Z.NA[M==1] <- NA
      }else if(marmech=="random"){
        Z.NA <- as.matrix(genMask_MAR(Z, mech = type, pmiss = pmiss))
        M <- Z.NA
        M[!is.na(M)] <- 0
        M[is.na(M)] <- 1
      }

    }else if (case==2){

      #(ii) Correlated Normal (Corr-N)
      Z<-gendist(n=n,p=p, dist="Gaussian", Sigma=Sigma, mu=rep(0,p))
      if(marmech =="basic"){
        M <- genMask(X = Z, n = n, p = p, r = r, type=  type)
        Z.NA <- Z
        Z.NA[M==1] <- NA
      }else if(marmech=="random"){
        Z.NA <- as.matrix(genMask_MAR(Z, mech = type, pmiss = pmiss))
        M <- Z.NA
        M[!is.na(M)] <- 0
        M[is.na(M)] <- 1
      }
    }else if (case==3){
      #(iii) Independent Multivariate t (t)
      Z<-gendist(n=n,p=p, dist="t", Sigma=diag(p), mu=rep(0,p), df=4)
      if(marmech =="basic"){
        M <- genMask(X = Z, n = n, p = p, r = r, type=  type)
        Z.NA <- Z
        Z.NA[M==1] <- NA
      }else if(marmech=="random"){
        Z.NA <- as.matrix(genMask_MAR(Z, mech = type, pmiss = pmiss))
        M <- Z.NA
        M[!is.na(M)] <- 0
        M[is.na(M)] <- 1
      }
    }else if(case==4){

      #(iv) Correlated Multivariate t (Corr-t)
      Z<-gendist(n=n,p=p, dist="t", Sigma=Sigma, mu=rep(0,p), df=4)
      if(marmech =="basic"){
        M <- genMask(X = Z, n = n, p = p, r = r, type=  type)
        Z.NA <- Z
        Z.NA[M==1] <- NA
      }else if(marmech=="random"){
        Z.NA <- as.matrix(genMask_MAR(Z, mech = type, pmiss = pmiss))
        M <- Z.NA
        M[!is.na(M)] <- 0
        M[is.na(M)] <- 1
      }
    }else if(case ==5){
      #(v) Independent Uniform distribution (U)
      Z<-gendist(n=n,p=p, dist="uniform", Sigma=diag(p))
      if(marmech =="basic"){
        M <- genMask(X = Z, n = n, p = p, r = r, type=  type)
        Z.NA <- Z
        Z.NA[M==1] <- NA
      }else if(marmech=="random"){
        Z.NA <- as.matrix(genMask_MAR(Z, mech = type, pmiss = pmiss))
        M <- Z.NA
        M[!is.na(M)] <- 0
        M[is.na(M)] <- 1
      }
    }else if (case==6){
      #(vi) Correlated Uniform distribution (Corr-U)
      Z<-gendist(n=n,p=p, dist="uniform", Sigma=Sigma)
      if(marmech =="basic"){
        M <- genMask(X = Z, n = n, p = p, r = r, type=  type)
        Z.NA <- Z
        Z.NA[M==1] <- NA
      }else if(marmech=="random"){
        Z.NA <- as.matrix(genMask_MAR(Z, mech = type, pmiss = pmiss))
        M <- Z.NA
        M[!is.na(M)] <- 0
        M[is.na(M)] <- 1
      }

    }else if(case ==7){
      #(vii) Weird W distribution (W)
      Z<-gendist(n=n,p=p, dist="W", Sigma=diag(p), mu=rep(0,p))
      if(marmech =="basic"){
        M <- genMask(X = Z, n = n, p = p, r = r, type=  type)
        Z.NA <- Z
        Z.NA[M==1] <- NA
      }else if(marmech=="random"){
        Z.NA <- as.matrix(genMask_MAR(Z, mech = type, pmiss = pmiss))
        M <- Z.NA
        M[!is.na(M)] <- 0
        M[is.na(M)] <- 1
      }
    }else if(case==8){

      #(viii) Independent Weibull distribution (Weibull)
      Z<-gendist(n=n,p=p, dist="Weibull", scale=1,shape=2)
      if(marmech =="basic"){
        M <- genMask(X = Z, n = n, p = p, r = r, type=  type)
        Z.NA <- Z
        Z.NA[M==1] <- NA
      }else if(marmech=="random"){
        Z.NA <- as.matrix(genMask_MAR(Z, mech = type, pmiss = pmiss))
        M <- Z.NA
        M[!is.na(M)] <- 0
        M[is.na(M)] <- 1
      }
      }else{
      Z<- genData_real(dataset=case, datapath=datapath)
      Z.NA <- as.matrix(genMask_MAR(Z, mech = type, pmiss = pmiss))
      M <- Z.NA
      M[!is.na(M)] <- 0
      M[is.na(M)] <- 1
      }


    if (runPKLMtest){
    start_time <- Sys.time()

    pvals.PKLMtest <- tryCatch({PKLMtest(X = Z.NA,
                             num.proj,
                             num.trees.per.proj = num.trees.per.proj,
                             nrep = nrep,
                             min.node.size,
                             size.resp.set=size.resp.set,
                             refit=refit)}, error=function(e) NA)
    end_time <- Sys.time()

    time.big <- as.numeric(end_time - start_time)
    }else{
      pvals.PKLMtest<-NA
      time.big<-NA
    }

    if (runlittleTest){
    start_time <- Sys.time()
    pvals.littletest <- tryCatch({mcar_test(as.data.frame(Z.NA))$p.value}, error=function(e) NA)

    end_time <- Sys.time()

    time.little <- as.numeric(end_time - start_time)
    }else{
      pvals.littletest<-NA
      time.little<-NA
    }
    
    if(runJJTest){
    start_time <- Sys.time()
    pvals.JJtest <- tryCatch({ JJtest(as.data.frame(Z.NA))$pnormality
    }, error = function(e) NA)

    end_time <- Sys.time()
    time.JJ <- as.numeric(end_time - start_time)
    }else{
      pvals.JJtest <- NA
      time.JJ <- NA  
    }

    if(runEnergyTest){
      start_time <- Sys.time()
      pvals.energyTest <- tryCatch({energyTest_original(X = Z.NA, nrep =nrep)}, error = function(e) NA)
      end_time <- Sys.time()
      time.energy <- as.numeric(end_time - start_time)
    }else{
      pvals.energyTest <- NA
      time.energy <- NA
    }
    print(paste("Big Test:", pvals.PKLMtest))
    print(paste("Energy Test:", pvals.energyTest))

    return(list(pvals.PKLMtest=pvals.PKLMtest, pvals.energyTest=pvals.energyTest,
                pvals.littletest=pvals.littletest, pvals.JJtest=pvals.JJtest,
                time.big= time.big, time.little=time.little, time.JJ=time.JJ, time.energy =time.energy))
  }, mc.cores = n.cores)



  return(pvals)
}



PKLMtest <- function(X,
                    num.proj = 20,
                    num.trees.per.proj = 200,
                    nrep = 20,
                    min.node.size = 10,
                    size.resp.set = 2,
                    compute.partial.pvals = FALSE,
                    refit=F,
                    ...) {
  
  # checks
  if (num.proj < 2) {
    stop("num-proj should be at least 2.")
  }
  
  # ensure that size.resp.set = numbers of collapsed classes
  size.resp.set <- size.resp.set - 1
  
  # checks
  if (!is.matrix(X)) {
    stop("X should be a matrix.")
  }
  
  # get the mask for the data provided in X
  M <- is.na(X)
  
  # copute the permutations of the missing value patterns
  Mperm <- lapply(1:nrep, function(j) M[sample(nrow(M)),])
  
  # list of results
  list.obs <- list()
  list.null <- list()
  list.var.cov <- list()
  list.var.resp <- list()
  
  # main loop
  while (length(list.obs) != num.proj) {
    
    
    # sample a projection for the response
    new.var <- c()
    num.classes <- 1
    while (length(new.var)==0 ||  num.classes <= 1 ) {
      
      # sample a projection for the covariates
      size.cov <- sample(1:ncol(X), size = 1, replace = FALSE)
      var.cov <- sample(1:ncol(X), size = size.cov, replace = FALSE)
      
      new.var <- setdiff(1:ncol(X), var.cov)
      
      if (length(new.var)>1) {
        
        if (size.resp.set=="random"){
          size.resp <- sample(1:length(new.var), size = 1, replace = FALSE)
        }else{
          size.resp<- sample(1:min(length(new.var), size.resp.set), size=1, replace=FALSE)
        }
        
        var.resp <- sample(new.var, size = size.resp, replace = FALSE)
      } else if (length(new.var)==1) {
        var.resp <- new.var
      } else if (length(new.var)==0){
        var.resp <- var.cov
      }
      
      # which one are complete there
      ids.keep <- which(apply(X[,var.cov,drop=FALSE],
                              1, function(x) !any(is.na(x))))
      
      # define the new response
      M.resp <- M[ids.keep,var.resp,drop=F]
      unique.pat.resp <- unique(M.resp)
      num.classes<- nrow(unique(M.resp))
      
    }
    
    list.var.cov[[length(list.var.cov)+1]] <- var.cov
    list.var.resp[[length(list.var.resp)+1]] <- var.resp
    
    
    ids.patterns.resp <- sapply(1:nrow(unique.pat.resp),
                                function(i) which(apply(M.resp,
                                                        1,
                                                        function(m) identical(m,
                                                                              unique.pat.resp[i,]))), simplify = F)
    
    # get the data for the projection
    X.proj <- X[ids.keep,var.cov,drop=F]
    
    # create the class for the projection
    class.resp <- rep(NA, length(ids.keep))
    
    for (i in 1:length(ids.patterns.resp)) {
      class.resp[ids.patterns.resp[[i]]] <- i
    }
    
    
    class.resp.perm<-lapply(1:nrep, FUN=function(j){
      
      
      M.respj <- Mperm[[j]][ids.keep,var.resp,drop=F]
      unique.pat.respj <- unique(M.respj)
      num.classesj<- nrow(unique(M.respj))
      
      
      ids.patterns.resp <- sapply(1:nrow(unique.pat.respj),
                                  function(i) which(apply(M.respj,
                                                          1,
                                                          function(m) identical(m,
                                                                                unique.pat.respj[i,]))), simplify = F)
      
      # create the class for the projection
      class.respj <- rep(NA, length(ids.keep))
      
      for (i in 1:length(ids.patterns.resp)) {
        class.respj[ids.patterns.resp[[i]]] <- i
      }
      
      return(class.respj)
      
      
    } )
    
    d <- data.frame(y = factor(class.resp),
                    X = X.proj)

    st <- tryCatch({ranger::ranger(y~., data = d,
                                   num.trees = num.trees.per.proj,
                                   classification = TRUE,
                                   min.node.size = min.node.size,
                                   probability = TRUE)}, error = function(e) NA)


    
    if(!any(is.na(st))) {
      
      # generate the observed stat
      

      obs <- genU(st, lab = class.resp)

      
      
      # generate the stat under H0
      null <- sapply(class.resp.perm, function(p){
        
   
        if (refit==T){
        
  
         d <- data.frame(y = factor(p),
                           X = X.proj)
        
         st <- ranger::ranger(y~., data = d,
                                        num.trees = num.trees.per.proj,
                                        classification = TRUE,
                                        min.node.size = min.node.size,
                                        probability = TRUE)

        return(genU(st, lab = p))
       
      
        
        }else{
           
          return(genU(st, lab = p))
          
          
          }
        })
      
      
      
      list.obs[[length(list.obs)+1]] <- obs
      list.null[[length(list.null)+1]] <- null
      
    }else{
      list.var.cov <- list.var.cov[-length(list.var.cov)]
      list.var.resp <- list.var.cov[-length(list.var.resp)]
    }
    
    
    
    
    
  }
  print("while done")
  
  stat.perm.raw <- Reduce(list.null,f = rbind)
  
  
  
  
  isIn <- sapply(1:ncol(X), function(i) sapply(list.var.resp, function(l) !(i %in% l))) # list.var.resp
  isIn <- cbind(rep(TRUE, num.proj ), isIn)
  if (!compute.partial.pvals) {
    isIn <- isIn[,1,drop=FALSE]
  }
  
  partial.perm <- apply(isIn ,2, function(x) {
    ids <- which(x)
    
    colMeans(stat.perm.raw[ids,,drop=FALSE], na.rm=T)
    
  })
  
  if (ncol(isIn)==1) {
    partial.perm <- matrix(partial.perm, nrow=1)
  }
  
  partial.obs <- apply(isIn ,2, function(x) {
    
    if (sum(x)==0) {
      return(NA)
    } else {
      ids <- which(x)
      return(mean(unlist(list.obs[ids]),
                  na.rm=T))
    }
    
    
  })
  
  ##### partial p-values #########
  
  if (compute.partial.pvals){
    pvals <- sapply(1:(ncol(X)+1), function(i) (sum(partial.perm[,i] >= partial.obs[i])+1)/(length(partial.perm[,i])+1))
    names(pvals) <- c("all", paste0("-",1:ncol(X)))
  }else{
    pvals<-(sum(partial.perm >= partial.obs)+1)/(length(partial.perm)+1)
    print(pvals)
    names(pvals) <- "all"
  }
  
  ##############################
  
  return(pvals)
  
}


expand.grid.unique <- function(x, y, include.equals=FALSE)
{
  x <- unique(x)

  y <- unique(y)

  g <- function(i)
  {
    z <- setdiff(y, x[seq_len(i-include.equals)])

    if(length(z)) cbind(x[i], z, deparse.level=0)
  }

  do.call(rbind, lapply(seq_along(x), g))
}



genMask_MAR <- function(X, mech = "MAR", pmiss = 0.5) {

  create.patterns <- function(pmiss, d){
    patterns <- c()
    nar <- 0
    while(nar != ceiling(d/2)){
      nr.of.0 <- rbinom(prob=pmiss,n = 1, size=d)
      if(nr.of.0 > ceiling(d/2)){
        nr.of.0 <- d-nr.of.0
      }

      pat <- rep(1, d)
      pat[sample(1:d, nr.of.0)]<-0
      patterns <- rbind(patterns, pat)
      patterns <- patterns[!duplicated(patterns), , drop=F]
      nar <- ifelse(is.null(nrow(patterns)),0,nrow(patterns))

    }

    return(patterns)
  }

  n <- nrow(X)
  d <- ncol(X)
  # method = simple deletes any entry in the data matrix with probability pNA
  if(mech == "MCAR"){
    na <- apply(X,2,function(x) sample(c(0,1), size = length(x), prob = c(1-pmiss, pmiss), replace = TRUE))
    X.NA <- X
    X.NA[na==1] <- NA
  }else if (mech %in% c("MAR", "MNAR")){

    res <- NA
    while(length(res)==1) {
      patterns <- create.patterns(pmiss,d)



      freq <- rep(1/nrow(patterns), nrow(patterns))
      cont <- TRUE
      bycases <- FALSE
      run <- TRUE

      res <- tryCatch(ampute(X,
                             prop = pmiss,
                             patterns = patterns,
                             freq = freq,
                             mech = mech,
                             cont = cont,
                             bycases = bycases,
                             run = run)$amp, error = function(x) NA)
    }
    X.NA <- res
  }


  return(X.NA)
}


######################JJTest##################################################
# This code is taken from the package "MissMech", see
#https://www.jstatsoft.org/article/view/v056i06

JJtest<- function (data, del.lesscases = 6, imputation.number = 1, method = "Auto",
                   imputation.method = "Dist.Free", nrep = 10000, n.min = 30,
                   seed = 110, alpha = 0.05, imputed.data = NA)
{
  if (!is.na(seed))
    set.seed(seed)
  if (is.data.frame(data)) {
    data <- as.matrix(data)
  }
  if (!is.na(imputed.data[1]) && imputation.number != 1) {
    cat("Warning: No multiple imputation allowed when imputed data is provided.\n")
  }
  if (!is.matrix(data)) {
    cat("Warning: Data is not a matrix or data frame.\n")
    stop("")
  }
  if (length(data) == 0) {
    cat("Warning: Data is empty.\n")
    stop("")
  }
  if (ncol(data) < 2) {
    cat("Warning: More than 1 variable is required.\n")
    stop("")
  }
  allempty <- which(apply(!is.na(data), 1, sum) == 0)
  if (length(allempty) != 0) {
    data <- data[apply(!is.na(data), 1, sum) != 0, ]
    cat("Warning:", length(allempty), "Cases with all variables missing have been removed \n\n          from the data.\n")
  }
  newdata <- OrderMissing(data, del.lesscases)
  if (length(newdata$data) == 0) {
    cat("Warning: There are no data sets after deleting insufficient cases.\n")
    stop("")
  }
  if (newdata$g == 1) {
    cat("Warning: More than one missing data pattern should be present.\n")
    stop("")
  }
  if (sum(newdata$patcnt == 1) > 0) {
    cat("Warning: At least 2 cases needed in each missing data patterns.\n")
    stop("")
  }
  y <- newdata$data
  patused <- newdata$patused
  patcnt <- newdata$patcnt
  spatcnt <- newdata$spatcnt
  caseorder <- newdata$caseorder
  removedcases <- newdata$removedcases
  n <- nrow(y)
  p <- ncol(y)
  g <- newdata$g
  spatcntz <- c(0, spatcnt)
  pvalsn <- matrix(0, imputation.number, g)
  adistar <- matrix(0, imputation.number, g)
  pnormality <- c()
  x <- vector("list", g)
  n4sim <- vector("list", g)
  mu <- matrix(0, p, 1)
  sig <- diag(1, p)
  emest <- Mls(newdata, mu, sig, 1e-06)
  mu <- emest$mu
  sig <- emest$sig
  if (is.na(imputed.data[1])) {
    yimp <- y
    if (imputation.method == "Dist.Free") {
      iscomp <- apply(patused, 1, sum, na.rm = TRUE) ==
        p
      cind <- which(iscomp)
      ncomp <- patcnt[cind]
      if (length(ncomp) == 0)
        ncomp <- 0
      use.normal <- FALSE
      if (ncomp >= 10 && ncomp >= 2 * p) {
        compy <- y[seq(spatcntz[cind] + 1, spatcntz[cind +
                                                      1]), ]
        ybar <- matrix(apply(compy, 2, mean))
        sbar <- cov(compy)
        resid <- (ncomp/(ncomp - 1))^0.5 * (compy - matrix(ybar,
                                                           ncomp, p, byrow = TRUE))
      }
      else {
        cat("Warning: There is not sufficient number of complete cases.\n  Dist.Free imputation requires a least 10 complete cases\n  or 2*number of variables, whichever is bigger.\n  imputation.method = normal will be used instead.\n")
        use.normal <- TRUE
      }
    }
    for (k in 1:imputation.number) {
      if (imputation.method == "Normal" || use.normal) {
        yimp <- Impute(data = y, mu, sig, imputation.method = "Normal")
        yimp <- yimp$yimpOrdered
      }
      if (imputation.method == "Dist.Free" && !use.normal) {
        yimp <- Impute(data = y, ybar, sbar, imputation.method = "Dist.Free",
                       resid)
        yimp <- yimp$yimpOrdered
      }
      if (k == 1)
        yimptemp <- yimp
      templist <- Hawkins(yimp, spatcnt)
      fij <- templist$fij
      tail <- templist$a
      ni <- templist$ni
      if (method == "Auto" || method == "Hawkins") {
        for (i in 1:g) {
          if (ni[i] < n.min) {
            if (k == 1) {
              n4sim[[i]] <- SimNey(ni[i], nrep)
            }
          }
          templist <- TestUNey(tail[[i]], nrep, sim = n4sim[[i]],
                               n.min)
          pn <- templist$pn
          n4 <- templist$n4
          pn <- pn + (pn == 0)/nrep
          pvalsn[k, i] <- pn
        }
      }
      if (method == "Auto" || method == "Nonparametric") {
        if (length(ni) < 2) {
          cat("Warning: Not enough groups for AndersonDarling test.")
          stop("")
        }
        templist <- AndersonDarling(fij, ni)
        p.ad <- templist$pn
        adistar[k, ] <- templist$adk.all
        pnormality <- c(pnormality, p.ad)
      }
    }
  }
  else {
    yimp <- imputed.data[caseorder, ]
    yimptemp <- yimp
    templist <- Hawkins(yimp, spatcnt)
    fij <- templist$fij
    tail <- templist$a
    ni <- templist$ni
    if (method == "Auto" || method == "Hawkins") {
      for (i in 1:g) {
        if (ni[i] < n.min) {
          n4sim[[i]] <- SimNey(ni[i], nrep)
        }
        templist <- TestUNey(tail[[i]], nrep, sim = n4sim[[i]],
                             n.min)
        pn <- templist$pn
        n4 <- templist$n4
        pn <- pn + (pn == 0)/nrep
        pvalsn[1, i] <- pn
      }
    }
    if (method == "Auto" || method == "Nonparametric") {
      templist <- AndersonDarling(fij, ni)
      p.ad <- templist$pn
      adistar[1, ] <- templist$adk.all
      pnormality <- c(pnormality, p.ad)
    }
  }
  adstar <- apply(adistar, 1, sum)
  combp <- -2 * apply(log(pvalsn), 1, sum)
  pvalcomb <- pchisq(combp, 2 * g, lower.tail = FALSE)
  if (method == "Hawkins") {
    pnormality <- NULL
    adstar <- NULL
    adistar <- NULL
  }
  if (method == "Nonparametric") {
    pvalcomb = NULL
    combp = NULL
    pvalsn = NULL
  }
  yimptemp <- yimptemp[order(caseorder), ]
  if (length(removedcases) == 0) {
    dataused <- data
  }
  else {
    dataused <- data[-1 * removedcases, ]
  }
  homoscedastic <- list(analyzed.data = dataused, imputed.data = yimptemp,
                        ordered.data = y, caseorder = caseorder, pnormality = pnormality,
                        adstar = adstar, adistar = adistar, pvalcomb = pvalcomb,
                        combp = combp, pvalsn = pvalsn, g = g, alpha = alpha,
                        patused = patused, patcnt = patcnt, imputation.number = imputation.number,
                        mu = mu, sigma = sig)
  homoscedastic$call <- match.call()
  class(homoscedastic) <- "testhomosc"
  homoscedastic
}

OrderMissing <- function (data, del.lesscases = 0)
{
  y <- data
  if (is.data.frame(y)) {
    y <- as.matrix(y)
  }
  if (!is.matrix(y)) {
    cat("Warning data is not a matrix or data frame")
    stop("")
  }
  if (length(y) == 0) {
    cat("Warning: data is empty")
    return
  }
  names <- colnames(y)
  n <- nrow(y)
  pp <- ncol(y)
  yfinal <- c()
  patused <- c()
  patcnt <- c()
  caseorder <- c()
  removedcases <- c()
  ordertemp <- c(1:n)
  ntemp <- n
  ptemp <- pp
  done <- FALSE
  yatone <- FALSE
  while (!done) {
    pattemp <- is.na(y[1, ])
    indin <- c()
    indout <- c()
    done <- TRUE
    for (i in 1:ntemp) {
      if (all(is.na(y[i, ]) == pattemp)) {
        indout <- c(indout, i)
      }
      else {
        indin <- c(indin, i)
        done <- FALSE
      }
    }
    if (length(indin) == 1)
      yatone = TRUE
    yfinal <- rbind(yfinal, y[indout, ])
    y <- y[indin, ]
    caseorder <- c(caseorder, ordertemp[indout])
    ordertemp <- ordertemp[indin]
    patcnt <- c(patcnt, length(indout))
    patused <- rbind(patused, pattemp)
    if (yatone) {
      pattemp <- is.na(y)
      yfinal <- rbind(yfinal, matrix(y, ncol = pp))
      y <- c()
      indin <- c()
      indout <- c(1)
      caseorder <- c(caseorder, ordertemp[indout])
      ordertemp <- ordertemp[indin]
      patcnt <- c(patcnt, length(indout))
      patused <- rbind(patused, pattemp)
      done <- TRUE
    }
    if (!done)
      ntemp <- nrow(y)
  }
  caseorder <- c(caseorder, ordertemp)
  patused <- ifelse(patused, NA, 1)
  rownames(patused) <- NULL
  colnames(patused) <- names
  spatcnt <- cumsum(patcnt)
  dataorder <- list(data = yfinal, patused = patused, patcnt = patcnt,
                    spatcnt = spatcnt, g = length(patcnt), caseorder = caseorder,
                    removedcases = removedcases)
  dataorder$call <- match.call()
  class(dataorder) <- "orderpattern"
  if (del.lesscases > 0) {
    dataorder <- DelLessData(dataorder, del.lesscases)
  }
  dataorder$patused <- matrix(dataorder$patused, ncol = pp)
  colnames(dataorder$patused) <- names
  dataorder
}


DelLessData<- function(data, ncases = 0)
{
  if (length(data) == 0) {
    cat("Warning: data is empty")
    return
  }
  if (is.matrix(data)) {
    data <- OrderMissing(data)
  }
  n <- nrow(data$data)
  p <- ncol(data$data)
  ind <- which(data$patcnt <= ncases)
  spatcntz <- c(0, data$spatcnt)
  rm <- c()
  removedcases <- c()
  if (length(ind) != 0) {
    for (i in 1:length(ind)) {
      rm <- c(rm, seq(spatcntz[ind[i]] + 1, spatcntz[ind[i] +
                                                       1]))
    }
    y <- data$data[-1 * rm, ]
    removedcases <- data$caseorder[rm]
    patused <- data$patused[-1 * ind, ]
    patcnt <- data$patcnt[-1 * ind]
    caseorder <- data$caseorder[-1 * rm]
    spatcnt <- cumsum(patcnt)
  }
  else {
    patused <- data$patused
    patcnt <- data$patcnt
    spatcnt <- data$spatcnt
    caseorder <- data$caseorder
    y <- data$data
  }
  newdata <- list(data = y, patused = patused, patcnt = patcnt,
                  spatcnt = spatcnt, g = length(patcnt), caseorder = caseorder,
                  removedcases = removedcases)
  class(newdata) <- "orderpattern"
  newdata
}


Mls <- function (data, mu = NA, sig = NA, tol = 1e-06, Hessian = FALSE)
{
  if (!is.matrix(data) && class(data) != "orderpattern") {
    cat("Warning: data must have the classes of matrix or orderpattern.\n")
    stop("")
  }
  if (is.matrix(data)) {
    allempty <- which(apply(!is.na(data), 1, sum) == 0)
    if (length(allempty) != 0) {
      data <- data[apply(!is.na(data), 1, sum) != 0, ]
      cat("Warning:", length(allempty), "Cases with all variables missing have been removed\n         from the data.\n")
    }
    data <- OrderMissing(data)
  }
  if (class(data) == "orderpattern") {
    allempty <- which(apply(!is.na(data$data), 1, sum) ==
                        0)
    if (length(allempty) != 0) {
      data <- data$data
      data <- data[apply(!is.na(data), 1, sum) != 0, ]
      cat("Warning:", length(allempty), "Cases with all variables missing have been removed\n         from the data.\n")
      data <- OrderMissing(data)
    }
  }
  if (length(data$data) == 0) {
    cat("Warning: Data is empty")
    stop("")
  }
  if (ncol(data$data) < 2) {
    cat("Warning: More than 1 variable is required.\n")
    stop("")
  }
  y <- data$data
  patused <- data$patused
  spatcnt <- data$spatcnt
  if (is.na(mu[1])) {
    mu <- matrix(0, ncol(y), 1)
    sig <- diag(1, ncol(y))
  }
  itcnt <- 0
  em <- 0
  repeat {
    emtemp <- Sexpect(y, mu, sig, patused, spatcnt)
    ysbar <- emtemp$ysbar
    sstar <- emtemp$sstar
    em <- max(abs(sstar - mu %*% t(mu) - sig), abs(mu - ysbar))
    mu <- ysbar
    sig <- sstar - mu %*% t(mu)
    itcnt <- itcnt + 1
    if (!(em > tol || itcnt < 2))
      (break)()
  }
  rownames(mu) <- colnames(y)
  colnames(sig) <- colnames(y)
  if (Hessian) {
    templist <- Ddf(y, mu, sig)
    hessian <- templist$dd
    stderror <- templist$se
    return(list(mu = mu, sig = sig, hessian = hessian, stderror = stderror,
                iteration = itcnt))
  }
  return(list(mu = mu, sig = sig, iteration = itcnt))
}

Sexpect <- function(y, mu, sig, patused, spatcnt)
{
  n <- nrow(y)
  pp <- ncol(y)
  sstar <- matrix(0, pp, pp)
  a <- nrow(mu)
  b <- ncol(mu)
  ysbar <- matrix(0, a, b)
  first <- 1
  for (i in 1:length(spatcnt)) {
    ni <- spatcnt[i] - first + 1
    stemp <- matrix(0, pp, pp)
    indm <- which(is.na(patused[i, ]))
    indo <- which(!is.na(patused[i, ]))
    yo <- matrix(y[first:spatcnt[i], indo], ni, length(indo))
    first <- spatcnt[i] + 1
    muo <- mu[indo]
    mum <- mu[indm]
    sigoo <- sig[indo, indo]
    sigooi <- solve(sigoo)
    soo <- t(yo) %*% yo
    stemp[indo, indo] <- soo
    ystemp <- matrix(0, ni, pp)
    ystemp[, indo] <- yo
    if (length(indm) != 0) {
      sigmo <- matrix(sig[indm, indo], length(indm), length(indo))
      sigmm <- sig[indm, indm]
      temp1 <- matrix(mum, ni, length(indm), byrow = TRUE)
      temp2 <- yo - matrix(muo, ni, length(indo), byrow = TRUE)
      ym <- temp1 + temp2 %*% sigooi %*% t(sigmo)
      som <- t(yo) %*% ym
      smm <- ni * (sigmm - sigmo %*% sigooi %*% t(sigmo)) +
        t(ym) %*% ym
      stemp[indo, indm] <- som
      stemp[indm, indo] <- t(som)
      stemp[indm, indm] <- smm
      ystemp[, indm] <- ym
    }
    sstar <- sstar + stemp
    if (ni == 1) {
      ysbar <- t(ystemp) + ysbar
    }
    else {
      ysbar <- apply(ystemp, 2, sum) + ysbar
    }
  }
  ysbar <- (1/n) * ysbar
  sstar <- (1/n) * sstar
  sstar <- (sstar + t(sstar))/2
  return(list(ysbar = ysbar, sstar = sstar))
}

Mimpute <- function (data, patused, mu, sig) {
  ni <- nrow(data)
  pp <- ncol(data)
  indm <- which(is.na(patused))
  indo <- which(!is.na(patused))
  pm <- length(indm)
  po <- length(indo)
  muo <- mu[indo]
  mum <- mu[indm]
  sigooi <- solve(sig[indo, indo])
  sigmo <- matrix(sig[indm, indo], pm, po)
  sigmm <- matrix(sig[indm, indm], pm, pm)
  ss1 <- sigmo %*% sigooi
  varymiss <- sigmm - ss1 %*% t(sigmo)
  expymiss <- matrix(mum, ni, pm, byrow = TRUE) + (data[, indo] -
                                                     matrix(muo, ni, po, byrow = TRUE)) %*% t(ss1)
  if (pm == 1) {
    a <- sqrt(varymiss)
  }
  else {
    svdvar <- svd(varymiss)
    a <- diag(sqrt(svdvar$d)) %*% t(svdvar$u)
  }
  data[, indm] = matrix(rnorm(ni * pm), ni, pm) %*% a + expymiss
  data
}



Impute <- function (data, mu = NA, sig = NA, imputation.method = "Normal",
          resid = NA)
{
  if (!is.matrix(data) && class(data) != "orderpattern") {
    cat("Warning: data must have the classes of matrix or orderpattern.\n")
    stop("")
  }
  if (is.matrix(data)) {
    allempty <- which(apply(!is.na(data), 1, sum) == 0)
    if (length(allempty) != 0) {
      data <- data[apply(!is.na(data), 1, sum) != 0, ]
      cat("Warning:", length(allempty), "Cases with all variables missing have been removed\n         from the data.\n")
    }
    data <- OrderMissing(data)
  }
  if (class(data) == "orderpattern") {
    allempty <- which(apply(!is.na(data$data), 1, sum) ==
                        0)
    if (length(allempty) != 0) {
      data <- data$data
      data <- data[apply(!is.na(data), 1, sum) != 0, ]
      cat("Warning:", length(allempty), "Cases with all variables missing have been removed\n         from the data.\n")
      data <- OrderMissing(data)
    }
  }
  if (length(data$data) == 0) {
    cat("Warning: data is empty")
    return
  }
  if (ncol(data$data) < 2) {
    cat("More than 1 variable is required.\n")
    stop("")
  }
  y <- data$data
  patused <- data$patused
  spatcnt <- data$spatcnt
  patcnt <- data$patcnt
  g <- data$g
  caseorder <- data$caseorder
  spatcntz <- c(0, spatcnt)
  p <- ncol(y)
  n <- nrow(y)
  yimp <- y
  use.normal <- TRUE
  if (imputation.method == "Dist.Free") {
    if (is.na(mu[1])) {
      ybar <- matrix(0, p, 1)
      sbar <- diag(1, p)
      iscomp <- apply(patused, 1, sum, na.rm = TRUE) ==
        p
      cind <- which(iscomp)
      ncomp <- patcnt[cind]
      use.normal <- FALSE
      if (ncomp >= 10 && ncomp >= 2 * p) {
        compy <- y[seq(spatcntz[cind] + 1, spatcntz[cind +
                                                      1]), ]
        ybar <- matrix(apply(compy, 2, mean))
        sbar <- cov(compy)
        if (is.na(resid[1])) {
          resid <- (ncomp/(ncomp - 1))^0.5 * (compy -
                                                matrix(ybar, ncomp, p, byrow = TRUE))
        }
      }
      else {
        cat("Warning: There is not sufficient number of complete cases.\n  Dist.free imputation requires a least 10 complete cases\n  or 2*number of variables, whichever is bigger.\n")
        return
      }
    }
    if (!is.na(mu[1])) {
      ybar <- mu
      sbar <- sig
      iscomp <- apply(patused, 1, sum, na.rm = TRUE) ==
        p
      cind <- which(iscomp)
      ncomp <- patcnt[cind]
      use.normal <- FALSE
      compy <- y[seq(spatcntz[cind] + 1, spatcntz[cind +
                                                    1]), ]
      if (is.na(resid[1])) {
        resid <- (ncomp/(ncomp - 1))^0.5 * (compy - matrix(ybar,
                                                           ncomp, p, byrow = TRUE))
      }
    }
    indsample <- sample(ncomp, n - ncomp, replace = TRUE)
    resstar <- resid[indsample, ]
    indres1 <- 1
    for (i in 1:g) {
      if (sum(patused[i, ], na.rm = TRUE) != p) {
        test <- y[(spatcntz[i] + 1):spatcntz[i + 1],
                  ]
        indres2 <- indres1 + patcnt[i] - 1
        test <- MimputeS(matrix(test, ncol = p), patused[i,
                                                         ], ybar, sbar, matrix(resstar[indres1:indres2,
                                                                                       ], ncol = p))
        indres1 <- indres2 + 1
        yimp[(spatcntz[i] + 1):spatcntz[i + 1], ] <- test
      }
    }
  }
  if (imputation.method == "Normal" | use.normal) {
    if (is.na(mu[1])) {
      emest <- Mls(data, tol = 1e-06)
      mu <- emest$mu
      sig <- emest$sig
    }
    for (i in 1:g) {
      if (sum(patused[i, ], na.rm = TRUE) != p) {
        test <- y[(spatcntz[i] + 1):spatcntz[i + 1],
                  ]
        test <- Mimpute(matrix(test, ncol = p), patused[i,], mu, sig)
        yimp[(spatcntz[i] + 1):spatcntz[i + 1], ] <- test
      }
    }
  }
  yimpord <- yimp
  yimp <- yimp[order(caseorder), ]
  imputed <- list(yimp = yimp, yimpOrdered = yimpord, caseorder = caseorder,
                  patused = patused, patcnt = patcnt)
  imputed
}


MimputeS <-function (data, patused, y1, s1, e)
{
  x <- data
  ni <- nrow(x)
  pp <- ncol(x)
  indm <- which(is.na(patused))
  indo <- which(!is.na(patused))
  pm <- length(indm)
  po <- length(indo)
  a <- matrix(s1[indm, indo], pm, po) %*% solve(s1[indo, indo])
  dif <- x[, indo] - matrix(y1[indo], ni, po, byrow = TRUE)
  z <- matrix(y1[indm], ni, pm, byrow = TRUE) + dif %*% t(a)
  etta <- matrix(e[, indm], ni, pm) - matrix(e[, indo], ni,
                                             po) %*% t(a)
  zij <- z + etta
  x[, indm] <- zij
  x
}


Hawkins<-function(data, spatcnt)
{
  y <- data
  n <- nrow(y)
  p <- ncol(y)
  g <- length(spatcnt)
  spool <- matrix(0, p, p)
  gind <- c(0, spatcnt)
  ygc <- matrix(0, n, p)
  ni <- matrix(0, g, 1)
  for (i in 1:g) {
    yg <- y[seq(gind[i] + 1, gind[i + 1]), ]
    ni[i] <- nrow(yg)
    spool <- spool + (ni[i] - 1) * cov(yg)
    ygmean <- apply(yg, 2, mean)
    ygc[seq(gind[i] + 1, gind[i + 1]), ] <- yg - matrix(ygmean,
                                                        ni[i], p, byrow = TRUE)
  }
  spool <- spool/(n - g)
  spool <- solve(spool)
  f <- matrix(0, n, 1)
  nu <- n - g - 1
  a <- vector("list", g)
  for (i in 1:g) {
    vij <- ygc[seq(gind[i] + 1, gind[i + 1]), ]
    vij <- apply(vij %*% spool * vij, 1, sum)
    vij <- vij * ni[i]
    f[seq(gind[i] + 1, gind[i + 1])] <- ((n - g - p) * vij)/(p *
                                                               ((ni[i] - 1) * (n - g) - vij))
    a[[i]] <- 1 - pf(f[seq(gind[i] + 1, gind[i + 1])], p,
                     (nu - p + 1))
  }
  list(fij = f, a = a, ni = ni)
}


TestUNey<- function(x, nrep = 10000, sim = NA, n.min = 30)
{
  n <- length(x)
  pi <- LegNorm(x)
  n4 <- (apply(pi$p1, 2, sum)^2 + apply(pi$p2, 2, sum)^2 +
           apply(pi$p3, 2, sum)^2 + apply(pi$p4, 2, sum)^2)/n
  if (n < n.min) {
    if (is.na(sim[1])) {
      sim <- SimNey(n, nrep)
    }
    pn <- length(which(sim > n4))/nrep
  }
  else {
    pn <- pchisq(n4, 4, lower.tail = FALSE)
  }
  list(pn = pn, n4 = n4)
}

LegNorm<-function (x)
{
  if (!is.matrix(x)) {
    x <- as.matrix(x)
  }
  x <- 2 * x - 1
  p0 <- matrix(1, nrow(x), ncol(x))
  p1 <- x
  p2 <- (3 * x * p1 - p0)/2
  p3 <- (5 * x * p2 - 2 * p1)/3
  p4 <- (7 * x * p3 - 3 * p2)/4
  p1 <- sqrt(3) * p1
  p2 <- sqrt(5) * p2
  p3 <- sqrt(7) * p3
  p4 <- 3 * p4
  list(p1 = p1, p2 = p2, p3 = p3, p4 = p4)
}

SimNey<-function (n, nrep)
{
  x <- matrix(runif(nrep * n), ncol = nrep)
  pi <- LegNorm(x)
  n4sim <- (apply(pi$p1, 2, sum)^2 + apply(pi$p2, 2, sum)^2 +
              apply(pi$p3, 2, sum)^2 + apply(pi$p4, 2, sum)^2)/n
  n4sim <- sort(n4sim)
}

AndersonDarling<-function(data, number.cases)
{
  x <- data
  ni <- number.cases
  if (length(ni) < 2) {
    cat("Warning: Not enough groups for AndersonDarling test.")
    stop("")
  }
  k <- length(ni)
  ni.z <- c(0, cumsum(ni))
  n <- length(x)
  x.sort <- sort(x)
  x.sort <- x.sort[1:(n - 1)]
  ind <- which(duplicated(x.sort) == 0)
  counts <- c(ind, length(x.sort) + 1) - c(0, ind)
  hj <- counts[2:(length(ind) + 1)]
  hn <- cumsum(hj)
  zj <- x.sort[ind]
  adk <- 0
  adk.all <- matrix(0, k, 1)
  for (i in 1:k) {
    ind <- (ni.z[i] + 1):ni.z[i + 1]
    templist <- expand.grid(zj, x[ind])
    b <- templist[, 1] == templist[, 2]
    fij <- apply(matrix(b, length(zj)), 1, sum)
    mij <- cumsum(fij)
    num <- (n * mij - ni[i] * hn)^2
    dem <- hn * (n - hn)
    adk.all[i] <- (1/ni[i] * sum(hj * (num/dem)))
    adk <- adk + adk.all[i]
  }
  adk <- (1/n) * adk
  adk.all <- adk.all/n
  j <- sum(1/ni)
  i <- seq(1:(n - 1))
  h <- sum(1/i)
  g <- 0
  for (i in 1:(n - 2)) {
    g <- g + (1/(n - i)) * sum(1/seq((i + 1), (n - 1)))
  }
  a <- (4 * g - 6) * (k - 1) + (10 - 6 * g) * j
  b <- (2 * g - 4) * k^2 + 8 * h * k + (2 * g - 14 * h - 4) *
    j - 8 * h + 4 * g - 6
  c <- (6 * h + 2 * g - 2) * k^2 + (4 * h - 4 * g + 6) * k +
    (2 * h - 6) * j + 4 * h
  d <- (2 * h + 6) * k^2 - 4 * h * k
  var.adk <- ((a * n^3) + (b * n^2) + (c * n) + d)/((n - 1) *
                                                      (n - 2) * (n - 3))
  if (var.adk < 0)
    var.adk = 0
  adk.s <- (adk - (k - 1))/sqrt(var.adk)
  a0 <- c(0.25, 0.1, 0.05, 0.025, 0.01)
  b0 <- c(0.675, 1.281, 1.645, 1.96, 2.326)
  b1 <- c(-0.245, 0.25, 0.678, 1.149, 1.822)
  b2 <- c(-0.105, -0.305, -0.362, -0.391, -0.396)
  c0 <- log((1 - a0)/a0)
  qnt <- b0 + b1/sqrt(k - 1) + b2/(k - 1)
  if (adk.s <= qnt[3]) {
    ind <- seq(1:4)
  }
  else {
    ind <- seq(2:5)
  }
  yy <- spline(qnt[ind], c0[ind], xout = adk.s)$y
  p <- 1/(1 + exp(yy))
  list(pn = p, adk.all = adk.all, adk = adk, var.sdk = var.adk)
}


